# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""Tests for physical functions."""
# pylint: disable=no-member, invalid-name
import pytest
import numpy as np

from astropy.modeling.physical_models import BlackBody, NFW
from astropy.modeling.fitting import LevMarLSQFitter

from astropy.tests.helper import assert_quantity_allclose
from astropy import units as u
from astropy.utils.exceptions import AstropyUserWarning
from astropy import cosmology
from astropy.utils.compat.optional_deps import HAS_SCIPY  # noqa


__doctest_skip__ = ["*"]


# BlackBody tests


@pytest.mark.parametrize("temperature", (3000 * u.K, 2726.85 * u.deg_C))
def test_blackbody_evaluate(temperature):

    b = BlackBody(temperature=temperature, scale=1.0)

    assert_quantity_allclose(b(1.4 * u.micron), 486787299458.15656 * u.MJy / u.sr)
    assert_quantity_allclose(b(214.13747 * u.THz), 486787299458.15656 * u.MJy / u.sr)


def test_blackbody_weins_law():
    b = BlackBody(293.0 * u.K)
    assert_quantity_allclose(b.lambda_max, 9.890006672986939 * u.micron)
    assert_quantity_allclose(b.nu_max, 17.22525080856469 * u.THz)


def test_blackbody_sefanboltzman_law():
    b = BlackBody(293.0 * u.K)
    assert_quantity_allclose(b.bolometric_flux, 133.02471751812573 * u.W / (u.m * u.m))


def test_blackbody_input_units():
    SLAM = u.erg / (u.cm ** 2 * u.s * u.AA * u.sr)
    SNU = u.erg / (u.cm ** 2 * u.s * u.Hz * u.sr)

    b_lam = BlackBody(3000*u.K, scale=1*SLAM)
    assert(b_lam.input_units['x'] == u.AA)

    b_nu = BlackBody(3000*u.K, scale=1*SNU)
    assert(b_nu.input_units['x'] == u.Hz)


def test_blackbody_return_units():
    # return of evaluate has no units when temperature has no units
    b = BlackBody(1000.0 * u.K, scale=1.0)
    assert not isinstance(b.evaluate(1.0 * u.micron, 1000.0, 1.0), u.Quantity)

    # return has "standard" units when scale has no units
    b = BlackBody(1000.0 * u.K, scale=1.0)
    assert isinstance(b(1.0 * u.micron), u.Quantity)
    assert b(1.0 * u.micron).unit == u.erg / (u.cm ** 2 * u.s * u.Hz * u.sr)

    # return has scale units when scale has units
    b = BlackBody(1000.0 * u.K, scale=1.0 * u.MJy / u.sr)
    assert isinstance(b(1.0 * u.micron), u.Quantity)
    assert b(1.0 * u.micron).unit == u.MJy / u.sr

    # scale has units but evaluate scale has no units
    assert_quantity_allclose(b.evaluate(1.0 * u.micron, 1000.0 * u.K, 4.0), 89668184.86321202 * u.MJy / u.sr)


@pytest.mark.skipif("not HAS_SCIPY")
def test_blackbody_fit():

    fitter = LevMarLSQFitter()

    b = BlackBody(3000 * u.K, scale=5e-17 * u.Jy / u.sr)

    wav = np.array([0.5, 5, 10]) * u.micron
    fnu = np.array([1, 10, 5]) * u.Jy / u.sr

    b_fit = fitter(b, wav, fnu, maxiter=1000)

    assert_quantity_allclose(b_fit.temperature, 2840.7438355865065 * u.K)
    assert_quantity_allclose(b_fit.scale, 5.803783292762381e-17)


def test_blackbody_overflow():
    """Test Planck function with overflow."""
    photlam = u.photon / (u.cm ** 2 * u.s * u.AA)
    wave = [0.0, 1000.0, 100000.0, 1e55]  # Angstrom
    temp = 10000.0  # Kelvin
    bb = BlackBody(temperature=temp * u.K, scale=1.0)
    with pytest.warns(
            AstropyUserWarning,
            match=r'Input contains invalid wavelength/frequency value\(s\)'):
        with np.errstate(all="ignore"):
            bb_lam = bb(wave) * u.sr
    flux = bb_lam.to(photlam, u.spectral_density(wave * u.AA)) / u.sr

    # First element is NaN, last element is very small, others normal
    assert np.isnan(flux[0])
    with np.errstate(all="ignore"):
        assert np.log10(flux[-1].value) < -134
    np.testing.assert_allclose(
        flux.value[1:-1], [0.00046368, 0.04636773], rtol=1e-3
    )  # 0.1% accuracy in PHOTLAM/sr
    with np.errstate(all="ignore"):
        flux = bb(1.0 * u.AA)
    assert flux.value == 0


def test_blackbody_exceptions_and_warnings():
    """Test exceptions."""

    # Negative temperature
    with pytest.raises(
            ValueError,
            match="Temperature should be positive: \\[-100.\\] K"):
        bb = BlackBody(-100 * u.K)
        bb(1.0 * u.micron)

    bb = BlackBody(5000 * u.K)

    # Zero wavelength given for conversion to Hz
    with pytest.warns(AstropyUserWarning, match='invalid') as w:
        bb(0 * u.AA)
    assert len(w) == 3  # 2 of these are RuntimeWarning from zero divide

    # Negative wavelength given for conversion to Hz
    with pytest.warns(AstropyUserWarning, match='invalid') as w:
        bb(-1.0 * u.AA)
    assert len(w) == 1

    # Test that a non surface brightness convertible scale unit raises an error
    with pytest.raises(
            ValueError,
            match="scale units not dimensionless or in surface brightness: Jy"):
        bb = BlackBody(5000 * u.K, scale=1.0 * u.Jy)


def test_blackbody_array_temperature():
    """Regression test to make sure that the temperature can be an array."""
    multibb = BlackBody([100, 200, 300] * u.K)
    flux = multibb(1.2 * u.mm)
    np.testing.assert_allclose(
        flux.value, [1.804908e-12, 3.721328e-12, 5.638513e-12], rtol=1e-5
    )

    flux = multibb([2, 4, 6] * u.mm)
    np.testing.assert_allclose(
        flux.value, [6.657915e-13, 3.420677e-13, 2.291897e-13], rtol=1e-5
    )

    multibb = BlackBody(np.ones(4) * u.K)
    flux = multibb(np.ones((3, 4)) * u.mm)
    assert flux.shape == (3, 4)


def test_blackbody_dimensionless():
    """Test support for dimensionless (but not unscaled) units for scale"""
    T = 3000 * u.K
    r = 1e14 * u.cm
    DL = 100 * u.Mpc
    scale = np.pi * (r / DL)**2

    bb1 = BlackBody(temperature=T, scale=scale)
    # even though we passed scale with units, we should be able to evaluate with unitless
    bb1.evaluate(0.5, T.value, scale.to_value(u.dimensionless_unscaled))

    bb2 = BlackBody(temperature=T, scale=scale.to_value(u.dimensionless_unscaled))
    bb2.evaluate(0.5, T.value, scale.to_value(u.dimensionless_unscaled))

    # bolometric flux for both cases should be equivalent
    assert(bb1.bolometric_flux == bb2.bolometric_flux)


@pytest.mark.skipif("not HAS_SCIPY")
def test_blackbody_dimensionless_fit():
    T = 3000 * u.K
    r = 1e14 * u.cm
    DL = 100 * u.Mpc
    scale = np.pi * (r / DL)**2

    bb1 = BlackBody(temperature=T, scale=scale)
    bb2 = BlackBody(temperature=T, scale=scale.to_value(u.dimensionless_unscaled))

    fitter = LevMarLSQFitter()

    wav = np.array([0.5, 5, 10]) * u.micron
    fnu = np.array([1, 10, 5]) * u.Jy / u.sr

    bb1_fit = fitter(bb1, wav, fnu, maxiter=1000)
    bb2_fit = fitter(bb2, wav, fnu, maxiter=1000)

    assert(bb1_fit.temperature == bb2_fit.temperature)


@pytest.mark.parametrize("mass", (2.0000000000000E15 * u.M_sun, 3.976819741e+45 * u.kg))
def test_NFW_evaluate(mass):
    """Evaluation, density, and radii validation of NFW model."""
    # Test parameters
    concentration = 8.5
    redshift = 0.63
    cosmo = cosmology.Planck15

    # Parsec tests
    # 200c Overdensity
    massfactor = ("critical", 200)

    n200c = NFW(mass=mass, concentration=concentration, redshift=redshift, cosmo=cosmo,
                massfactor=massfactor)
    assert_quantity_allclose(n200c(3.0 * u.Mpc), (3.709693508e+12 * (u.solMass / u.Mpc ** 3),
                                                  7.376391187e+42 * (u.kg / u.Mpc ** 3)))
    assert_quantity_allclose(n200c.rho_scale, (7800150779863018.0 * (u.solMass / u.Mpc ** 3)))
    assert_quantity_allclose(n200c.r_s, (0.24684627641195428 * u.Mpc))
    assert_quantity_allclose(n200c.r_virial, (2.0981933495016114 * u.Mpc))

    # 200m Overdensity
    massfactor = ("mean", 200)

    n200m = NFW(mass=mass, concentration=concentration, redshift=redshift, cosmo=cosmo,
                massfactor=massfactor)
    assert_quantity_allclose(n200m(3.0 * u.Mpc), (3.626093406e+12 * (u.solMass / u.Mpc**3),
                                                  7.210159921e+42 * (u.kg / u.Mpc**3)))
    assert_quantity_allclose(n200m.rho_scale, (5118547639858115.0 * (u.solMass / u.Mpc ** 3)))
    assert_quantity_allclose(n200m.r_s, (0.2840612517326848 * u.Mpc))
    assert_quantity_allclose(n200m.r_virial, (2.414520639727821 * u.Mpc))

    # Virial mass
    massfactor = ("virial")

    nvir = NFW(mass=mass, concentration=concentration, redshift=redshift, cosmo=cosmo,
               massfactor=massfactor)
    assert_quantity_allclose(nvir(3.0 * u.Mpc), (3.646475546e+12 * (u.solMass / u.Mpc**3),
                                                 7.250687967e+42 * (u.kg / u.Mpc**3)))
    assert_quantity_allclose(nvir.rho_scale, (5649367524651067.0 * (u.solMass / u.Mpc ** 3)))
    assert_quantity_allclose(nvir.r_s, (0.2748701862303786 * u.Mpc))
    assert_quantity_allclose(nvir.r_virial, (2.3363965829582183 * u.Mpc))

    # kpc tests
    # 200c Overdensity
    massfactor = ("critical", 200)

    n200c = NFW(mass=mass, concentration=concentration, redshift=redshift, cosmo=cosmo,
                massfactor=massfactor)
    assert_quantity_allclose(n200c(3141 * u.kpc), (3254.373619264334 * (u.solMass / u.kpc ** 3),
                                                   6.471028627484543e+33 * (u.kg / u.kpc ** 3)))
    assert_quantity_allclose(n200c.rho_scale, (7800150.779863021 * (u.solMass / u.kpc ** 3)))
    assert_quantity_allclose(n200c.r_s, (246.84627641195425 * u.kpc))
    assert_quantity_allclose(n200c.r_virial, (2098.193349501611 * u.kpc))

    # 200m Overdensity
    massfactor = ("mean", 200)

    n200m = NFW(mass=mass, concentration=concentration, redshift=redshift, cosmo=cosmo,
                massfactor=massfactor)
    assert_quantity_allclose(n200m(3141 * u.kpc), (3184.0370866188623 * (u.solMass / u.kpc**3),
                                                   6.33117077170161e+33 * (u.kg / u.kpc**3)))
    assert_quantity_allclose(n200m.rho_scale, (5118547.639858116 * (u.solMass / u.kpc ** 3)))
    assert_quantity_allclose(n200m.r_s, (284.0612517326848 * u.kpc))
    assert_quantity_allclose(n200m.r_virial, (2414.5206397278207 * u.kpc))

    # Virial mass
    massfactor = ("virial")

    nvir = NFW(mass=mass, concentration=concentration, redshift=redshift, cosmo=cosmo,
               massfactor=massfactor)
    assert_quantity_allclose(nvir(3141 * u.kpc), (3201.1946851294997 * (u.solMass / u.kpc**3),
                                                  6.365287109937637e+33 * (u.kg / u.kpc**3)))
    assert_quantity_allclose(nvir.rho_scale, (5649367.5246510655 * (u.solMass / u.kpc ** 3)))
    assert_quantity_allclose(nvir.r_s, (274.87018623037864 * u.kpc))
    assert_quantity_allclose(nvir.r_virial, (2336.3965829582185 * u.kpc))

    # Meter tests
    # 200c Overdensity
    massfactor = ("critical", 200)

    n200c = NFW(mass=mass, concentration=concentration, redshift=redshift, cosmo=cosmo,
                massfactor=massfactor)
    assert_quantity_allclose(n200c(4.2e+23 * u.m), (1.527649658673012e-57 * (u.solMass / u.m ** 3),
                                                    3.0375936602739256e-27 * (u.kg / u.m ** 3)))
    assert_quantity_allclose(n200c.rho_scale, (2.654919529637763e-52 * (u.solMass / u.m ** 3)))
    assert_quantity_allclose(n200c.r_s, (7.616880211930209e+21 * u.m))
    assert_quantity_allclose(n200c.r_virial, (6.474348180140678e+22 * u.m))

    # 200m Overdensity
    massfactor = ("mean", 200)

    n200m = NFW(mass=mass, concentration=concentration, redshift=redshift, cosmo=cosmo,
                massfactor=massfactor)
    assert_quantity_allclose(n200m(4.2e+23 * u.m), (1.5194778058079436e-57 * (u.solMass / u.m ** 3),
                                                    3.0213446673751314e-27 * (u.kg / u.m ** 3)))
    assert_quantity_allclose(n200m.rho_scale, (1.742188385322371e-52 * (u.solMass / u.m ** 3)))
    assert_quantity_allclose(n200m.r_s, (8.76521436235054e+21 * u.m))
    assert_quantity_allclose(n200m.r_virial, (7.450432207997959e+22 * u.m))

    # Virial mass
    massfactor = ("virial")

    nvir = NFW(mass=mass, concentration=concentration, redshift=redshift, cosmo=cosmo,
               massfactor=massfactor)
    assert_quantity_allclose(nvir(4.2e+23 * u.m), (1.5214899184117633e-57 * (u.solMass / u.m ** 3),
                                                   3.0253455719375224e-27 * (u.kg / u.m ** 3)))
    assert_quantity_allclose(nvir.rho_scale, (1.922862338766335e-52 * (u.solMass / u.m ** 3)))
    assert_quantity_allclose(nvir.r_s, (8.481607714647913e+21 * u.m))
    assert_quantity_allclose(nvir.r_virial, (7.209366557450727e+22 * u.m))

    # Verify string input of overdensity type
    # 200c Overdensity
    massfactor = "200c"

    n200c = NFW(mass=mass, concentration=concentration, redshift=redshift, cosmo=cosmo,
                massfactor=massfactor)
    assert_quantity_allclose(n200c(3.0 * u.Mpc), (3.709693508e+12 * (u.solMass / u.Mpc ** 3),
                                                  7.376391187e+42 * (u.kg / u.Mpc ** 3)))

    # 200m Overdensity
    massfactor = "200m"

    n200m = NFW(mass=mass, concentration=concentration, redshift=redshift, cosmo=cosmo,
                massfactor=massfactor)
    assert_quantity_allclose(n200m(3.0 * u.Mpc), (3.626093406e+12 * (u.solMass / u.Mpc**3),
                                                  7.210159921e+42 * (u.kg / u.Mpc**3)))

    # Virial mass
    massfactor = "virial"

    nvir = NFW(mass=mass, concentration=concentration, redshift=redshift,  cosmo=cosmo,
               massfactor=massfactor)
    assert_quantity_allclose(nvir(3.0 * u.Mpc), (3.646475546e+12 * (u.solMass / u.Mpc**3),
                                                 7.250687967e+42 * (u.kg / u.Mpc**3)))


@pytest.mark.skipif("not HAS_SCIPY")
def test_NFW_fit():
    """Test linear fitting of NFW model."""
    # Fixed parameters
    redshift = 0.63
    cosmo = cosmology.Planck15

    # Radial set
    r = np.array([1.00e+01, 1.00e+02, 2.00e+02, 2.50e+02, 3.00e+02, 4.00e+02, 5.00e+02,
                  7.50e+02, 1.00e+03, 1.50e+03, 2.50e+03, 6.50e+03, 1.15e+04]) * u.kpc

    # 200c Overdensity
    massfactor = ("critical", 200)

    density_r = np.array([1.77842761e+08, 9.75233623e+06, 2.93789626e+06, 1.90107238e+06,
                          1.30776878e+06, 7.01004140e+05, 4.20678479e+05, 1.57421880e+05,
                          7.54669701e+04, 2.56319769e+04, 6.21976562e+03, 3.96522424e+02,
                          7.39336808e+01]) * (u.solMass / u.kpc ** 3)

    fitter = LevMarLSQFitter()

    n200c = NFW(mass=1.8E15 * u.M_sun, concentration=7.0, redshift=redshift, cosmo=cosmo,
                massfactor=massfactor)
    n200c.redshift.fixed = True

    n_fit = fitter(n200c, r, density_r, maxiter=1000)

    assert_quantity_allclose(n_fit.mass, 2.0000000000000E15 * u.M_sun)
    assert_quantity_allclose(n_fit.concentration, 8.5)

    # 200m Overdensity
    massfactor = ("mean", 200)

    density_r = np.array([1.35677282e+08, 7.95392979e+06, 2.50352599e+06, 1.64535870e+06,
                          1.14642248e+06, 6.26805453e+05, 3.81691731e+05, 1.46294819e+05,
                          7.11559560e+04, 2.45737796e+04, 6.05459585e+03, 3.92183991e+02,
                          7.34674416e+01]) * (u.solMass / u.kpc ** 3)

    fitter = LevMarLSQFitter()

    n200m = NFW(mass=1.8E15 * u.M_sun, concentration=7.0, redshift=redshift, cosmo=cosmo,
                massfactor=massfactor)
    n200m.redshift.fixed = True

    n_fit = fitter(n200m, r, density_r, maxiter=1000)

    assert_quantity_allclose(n_fit.mass, 2.0000000000000E15 * u.M_sun)
    assert_quantity_allclose(n_fit.concentration, 8.5)

    # Virial mass
    massfactor = ("virial", 200)

    density_r = np.array([1.44573515e+08, 8.34873998e+06, 2.60137484e+06, 1.70348738e+06,
                          1.18337370e+06, 6.43994654e+05, 3.90800249e+05, 1.48930537e+05,
                          7.21856397e+04, 2.48289464e+04, 6.09477095e+03, 3.93248818e+02,
                          7.35821787e+01]) * (u.solMass / u.kpc ** 3)

    fitter = LevMarLSQFitter()

    nvir = NFW(mass=1.8E15 * u.M_sun, concentration=7.0, redshift=redshift, cosmo=cosmo,
               massfactor=massfactor)
    nvir.redshift.fixed = True

    n_fit = fitter(nvir, r, density_r, maxiter=1000)

    assert_quantity_allclose(n_fit.mass, 2.0000000000000E15 * u.M_sun)
    assert_quantity_allclose(n_fit.concentration, 8.5)


def test_NFW_circular_velocity():
    """Test circular velocity and radial validation of NFW model."""
    # Test parameters
    mass = 2.0000000000000E15 * u.M_sun
    concentration = 8.5
    redshift = 0.63
    cosmo = cosmology.Planck15

    r_r = np.array([0.01, 0.1, 0.2, 0.25, 0.3, 0.4, 0.5, 0.75, 1.0, 1.5, 2.5, 6.5, 11.5]) * u.Mpc

    # 200c Overdensity tests
    massfactor = ("critical", 200)

    n200c = NFW(mass=mass, concentration=concentration, redshift=redshift, cosmo=cosmo,
                massfactor=massfactor)
    circ_v_200c = np.array([702.45487454, 1812.4138346, 2150.50929296, 2231.5802568, 2283.96950242,
                            2338.45989696, 2355.78876772, 2332.41766543, 2276.89433811,
                            2154.53909153, 1950.07947819, 1512.37442943,
                            1260.94034541]) * (u.km / u.s)
    assert_quantity_allclose(n200c.circular_velocity(r_r), circ_v_200c)
    assert_quantity_allclose(n200c.r_max, (0.5338248204429641 * u.Mpc))
    assert_quantity_allclose(n200c.v_max, (2356.7204380904027 * (u.km / u.s)))

    # 200m Overdensity tests
    massfactor = ("mean", 200)
    mass = 1.0e14 * u.M_sun
    concentration = 12.3
    redshift = 1.5

    n200m = NFW(mass=mass, concentration=concentration, redshift=redshift, cosmo=cosmo,
                massfactor=massfactor)
    circ_v_200m = np.array([670.18236647, 1088.9843324, 1046.82334367, 1016.88890732, 987.97273478,
                            936.00207134, 891.80115232, 806.63307977, 744.91002191, 659.33401039,
                            557.82823549, 395.9735786, 318.29863006]) * (u.km / u.s)
    assert_quantity_allclose(n200m.circular_velocity(r_r), circ_v_200m)
    assert_quantity_allclose(n200m.r_max, (0.10196917920081808 * u.Mpc))
    assert_quantity_allclose(n200m.v_max, (1089.0224395818727 * (u.km / u.s)))

    # Virial Overdensity tests
    massfactor = ("virial")
    mass = 1.2e+45 * u.kg
    concentration = 2.4
    redshift = 0.34

    r_r = np.array([3.08567758e+20, 3.08567758e+21, 6.17135516e+21, 7.71419395e+21,
                    9.25703274e+21, 1.23427103e+22, 1.54283879e+22, 2.31425819e+22,
                    3.08567758e+22, 4.62851637e+22, 7.71419395e+22, 2.00569043e+23,
                    3.54852922e+23]) * u.m

    nvir = NFW(mass=mass, concentration=concentration, redshift=redshift, cosmo=cosmo,
               massfactor=massfactor)
    circ_v_vir = np.array([205.87461783, 604.65091823, 793.9190629, 857.52516521, 908.90280843,
                           986.53582718, 1041.69089845, 1124.19719446, 1164.58270747, 1191.33193561,
                           1174.02934755, 1023.69360527, 895.52206321]) * (u.km / u.s)
    assert_quantity_allclose(nvir.circular_velocity(r_r), circ_v_vir)
    assert_quantity_allclose(nvir.r_max, (1.6484542328623448 * u.Mpc))
    assert_quantity_allclose(nvir.v_max, (1192.3130989914962 * (u.km / u.s)))


def test_NFW_exceptions_and_warnings_and_misc():
    """Test NFW exceptions."""

    # Arbitrary Test parameters
    mass = 2.0000000000000E15 * u.M_sun
    concentration = 8.5
    redshift = 0.63
    cosmo = cosmology.Planck15
    massfactor = ("critical", 200)

    r_r = np.array([1.00e+01, 1.00e+02, 2.00e+02, 2.50e+02, 3.00e+02, 4.00e+02, 5.00e+02,
                    7.50e+02, 1.00e+03, 1.50e+03, 2.50e+03, 6.50e+03, 1.15e+04]) * u.kpc

    # Massfactor exception tests
    with pytest.raises(ValueError) as exc:
        NFW(mass=mass, concentration=concentration, redshift=redshift, cosmo=cosmo,
            massfactor=("not", "virial"))
    assert exc.value.args[0] == "Massfactor 'not' not one of 'critical', 'mean', or 'virial'"
    with pytest.raises(ValueError) as exc:
        NFW(mass=mass, concentration=concentration, redshift=redshift, cosmo=cosmo,
            massfactor="not virial")
    assert exc.value.args[0] == "Massfactor not virial string not of the form '#m', '#c', " \
                                "or 'virial'"
    with pytest.raises(TypeError) as exc:
        NFW(mass=mass, concentration=concentration, redshift=redshift, cosmo=cosmo,
            massfactor=200)
    assert exc.value.args[0] == "Massfactor 200 not a tuple or string"

    # Verify unitless mass
    # Density test
    n200c = NFW(mass=mass.value, concentration=concentration, redshift=redshift, cosmo=cosmo,
                massfactor=massfactor)
    assert_quantity_allclose(n200c(3000.0), (3.709693508e+12 * (u.solMass / u.Mpc ** 3),
                                             7.376391187e+42 * (u.kg / u.Mpc ** 3)))

    # Circular velocity test with unitless mass
    circ_v_200c = np.array([702.45487454, 1812.4138346, 2150.50929296, 2231.5802568, 2283.96950242,
                            2338.45989696, 2355.78876772, 2332.41766543, 2276.89433811,
                            2154.53909153, 1950.07947819, 1512.37442943,
                            1260.94034541]) * (u.km / u.s)
    assert_quantity_allclose(n200c.circular_velocity(r_r), circ_v_200c)
    # test with unitless input velocity
    assert_quantity_allclose(n200c.circular_velocity(r_r.value), circ_v_200c)

    # Test Default Cosmology
    ncos = NFW(mass=mass, concentration=concentration, redshift=redshift)
    assert_quantity_allclose(ncos.A_NFW(concentration), 1.356554956501232)



# Licensed under a 3-clause BSD style license - see LICENSE.rst

import math
import numpy as np
import pytest

from astropy.convolution.convolve import convolve, convolve_fft, convolve_models
from astropy.modeling import models, fitting
from astropy.utils.misc import NumpyRNGContext
from astropy.utils.compat.optional_deps import HAS_SCIPY  # noqa
from numpy.testing import assert_allclose, assert_almost_equal


class TestConvolve1DModels:
    @pytest.mark.parametrize('mode', ['convolve_fft', 'convolve'])
    @pytest.mark.skipif('not HAS_SCIPY')
    def test_is_consistency_with_astropy_convolution(self, mode):
        kernel = models.Gaussian1D(1, 0, 1)
        model = models.Gaussian1D(1, 0, 1)
        model_conv = convolve_models(model, kernel, mode=mode)
        x = np.arange(-5, 6)
        ans = eval(f"{mode}(model(x), kernel(x))")

        assert_allclose(ans, model_conv(x), atol=1e-5)

    @pytest.mark.parametrize('mode', ['convolve_fft', 'convolve'])
    @pytest.mark.skipif('not HAS_SCIPY')
    def test_against_scipy(self, mode):
        from scipy.signal import fftconvolve

        kernel = models.Gaussian1D(1, 0, 1)
        model = models.Gaussian1D(1, 0, 1)
        model_conv = convolve_models(model, kernel, mode=mode)
        x = np.arange(-5, 6)
        ans = fftconvolve(kernel(x), model(x), mode='same')

        assert_allclose(ans, model_conv(x) * kernel(x).sum(), atol=1e-5)

    @pytest.mark.parametrize('mode', ['convolve_fft', 'convolve'])
    @pytest.mark.skipif('not HAS_SCIPY')
    def test_against_scipy_with_additional_keywords(self, mode):
        from scipy.signal import fftconvolve

        kernel = models.Gaussian1D(1, 0, 1)
        model = models.Gaussian1D(1, 0, 1)
        model_conv = convolve_models(model, kernel, mode=mode,
                                     normalize_kernel=False)
        x = np.arange(-5, 6)
        ans = fftconvolve(kernel(x), model(x), mode='same')

        assert_allclose(ans, model_conv(x), atol=1e-5)

    @pytest.mark.parametrize('mode', ['convolve_fft', 'convolve'])
    def test_sum_of_gaussians(self, mode):
        """
        Test that convolving N(a, b) with N(c, d) gives N(a + c, b + d),
        where N(., .) stands for Gaussian probability density function,
        in which a and c are their means and b and d are their variances.
        """

        kernel = models.Gaussian1D(1 / math.sqrt(2 * np.pi), 1, 1)
        model = models.Gaussian1D(1 / math.sqrt(2 * np.pi), 3, 1)
        model_conv = convolve_models(model, kernel, mode=mode,
                                     normalize_kernel=False)
        ans = models.Gaussian1D(1 / (2 * math.sqrt(np.pi)), 4, np.sqrt(2))
        x = np.arange(-5, 6)

        assert_allclose(ans(x), model_conv(x), atol=1e-3)

    @pytest.mark.parametrize('mode', ['convolve_fft', 'convolve'])
    def test_convolve_box_models(self, mode):
        kernel = models.Box1D()
        model = models.Box1D()
        model_conv = convolve_models(model, kernel, mode=mode)
        x = np.linspace(-1, 1, 99)
        ans = (x + 1) * (x < 0) + (-x + 1) * (x >= 0)

        assert_allclose(ans, model_conv(x), atol=1e-3)

    @pytest.mark.parametrize('mode', ['convolve_fft', 'convolve'])
    @pytest.mark.skipif('not HAS_SCIPY')
    def test_fitting_convolve_models(self, mode):
        """
        test that a convolve model can be fitted
        """
        b1 = models.Box1D()
        g1 = models.Gaussian1D()

        x = np.linspace(-5, 5, 99)
        fake_model = models.Gaussian1D(amplitude=10)
        with NumpyRNGContext(123):
            fake_data = fake_model(x) + np.random.normal(size=len(x))

        init_model = convolve_models(b1, g1, mode=mode, normalize_kernel=False)
        fitter = fitting.LevMarLSQFitter()
        fitted_model = fitter(init_model, x, fake_data)

        me = np.mean(fitted_model(x) - fake_data)
        assert_almost_equal(me, 0.0, decimal=2)



# Licensed under a 3-clause BSD style license - see LICENSE.rst
import pytest

import numpy as np

from astropy import wcs
from astropy.io import fits


class SimModelTAB:
    def __init__(self, nx=150, ny=200, crpix=[1, 1], crval = [1, 1],
                 cdelt = [1, 1], pc = {'PC1_1': 1, 'PC2_2': 1}):
        """  set essential parameters of the model (coord transformations)  """
        assert nx > 2 and ny > 1  # a limitation of this particular simulation
        self.nx = nx
        self.ny = ny
        self.crpix = crpix
        self.crval = crval
        self.cdelt = cdelt
        self.pc = pc

    def fwd_eval(self, xy):
        xb = 1 + self.nx // 3
        px = np.array([1, xb, xb, self.nx + 1])
        py = np.array([1, self.ny + 1])

        xi = self.crval[0] + self.cdelt[0] * (px - self.crpix[0])
        yi = self.crval[1] + self.cdelt[1] * (py - self.crpix[1])

        cx = np.array([0.0, 0.26, 0.8, 1.0])
        cy = np.array([-0.5, 0.5])

        xy = np.atleast_2d(xy)
        x = xy[:, 0]
        y = xy[:, 1]

        mbad = (x < px[0]) | (y < py[0]) | (x > px[-1]) | (y > py[-1])
        mgood = np.logical_not(mbad)

        i = 2 * (x > xb).astype(int)

        psix = self.crval[0] + self.cdelt[0] * (x - self.crpix[0])
        psiy = self.crval[1] + self.cdelt[1] * (y - self.crpix[1])

        cfx = (psix - xi[i]) / (xi[i + 1] - xi[i])
        cfy = (psiy - yi[0]) / (yi[1] - yi[0])

        ra = cx[i] + cfx * (cx[i + 1] - cx[i])
        dec = cy[0] + cfy * (cy[1] - cy[0])

        return np.dstack([ra, dec])[0]

    @property
    def hdulist(self):
        """ Simulates 2D data with a _spatial_ WCS that uses the ``-TAB``
        algorithm with indexing.
        """
        # coordinate array (some "arbitrary" numbers with a "jump" along x axis):
        x = np.array([[0.0, 0.26, 0.8, 1.0], [0.0, 0.26, 0.8, 1.0]])
        y = np.array([[-0.5, -0.5, -0.5, -0.5], [0.5, 0.5, 0.5, 0.5]])
        c = np.dstack([x, y])

        # index arrays (skip PC matrix for simplicity - assume it is an
        # identity matrix):
        xb = 1 + self.nx // 3
        px = np.array([1, xb, xb, self.nx + 1])
        py = np.array([1, self.ny + 1])
        xi = self.crval[0] + self.cdelt[0] * (px - self.crpix[0])
        yi = self.crval[1] + self.cdelt[1] * (py - self.crpix[1])

        # structured array (data) for binary table HDU:
        arr = np.array(
            [(c, xi, yi)],
            dtype=[
                ('wavelength', np.float64, c.shape),
                ('xi', np.double, (xi.size,)),
                ('yi', np.double, (yi.size,))
            ]
        )

        # create binary table HDU:
        bt = fits.BinTableHDU(arr);
        bt.header['EXTNAME'] = 'WCS-TABLE'

        # create primary header:
        image_data = np.ones((self.ny, self.nx), dtype=np.float32)
        pu = fits.PrimaryHDU(image_data)
        pu.header['ctype1'] = 'RA---TAB'
        pu.header['ctype2'] = 'DEC--TAB'
        pu.header['naxis1'] = self.nx
        pu.header['naxis2'] = self.ny
        pu.header['PS1_0'] = 'WCS-TABLE'
        pu.header['PS2_0'] = 'WCS-TABLE'
        pu.header['PS1_1'] = 'wavelength'
        pu.header['PS2_1'] = 'wavelength'
        pu.header['PV1_3'] = 1
        pu.header['PV2_3'] = 2
        pu.header['CUNIT1'] = 'deg'
        pu.header['CUNIT2'] = 'deg'
        pu.header['CDELT1'] = self.cdelt[0]
        pu.header['CDELT2'] = self.cdelt[1]
        pu.header['CRPIX1'] = self.crpix[0]
        pu.header['CRPIX2'] = self.crpix[1]
        pu.header['CRVAL1'] = self.crval[0]
        pu.header['CRVAL2'] = self.crval[1]
        pu.header['PS1_2'] = 'xi'
        pu.header['PS2_2'] = 'yi'
        for k, v in self.pc.items():
            pu.header[k] = v

        hdulist = fits.HDUList([pu, bt])
        return hdulist



from astropy.tests.helper import assert_quantity_allclose
from astropy.time import Time
from astropy.coordinates.builtin_frames.utils import get_polar_motion, get_offset_sun_from_barycenter
from astropy.coordinates.solar_system import get_body_barycentric_posvel
from astropy.utils.exceptions import AstropyWarning
import pytest


def test_polar_motion_unsupported_dates():
    msg = r'Tried to get polar motions for times {} IERS.*'

    with pytest.warns(AstropyWarning, match=msg.format('before')):
        get_polar_motion(Time('1900-01-01'))

    with pytest.warns(AstropyWarning, match=msg.format('after')):
        get_polar_motion(Time('2100-01-01'))


def test_sun_from_barycenter_offset():
    time = Time('2020-01-01')
    pos, vel = get_body_barycentric_posvel('sun', time)

    offset = get_offset_sun_from_barycenter(time)
    assert_quantity_allclose(offset.xyz, pos.xyz)
    assert not bool(offset.differentials)

    offset_with_vel = get_offset_sun_from_barycenter(time, include_velocity=True)
    assert_quantity_allclose(offset_with_vel.xyz, pos.xyz)
    assert_quantity_allclose(offset_with_vel.differentials['s'].d_xyz, vel.xyz)

    reverse = get_offset_sun_from_barycenter(time, reverse=True)
    assert_quantity_allclose(reverse.xyz, -pos.xyz)
    assert not bool(reverse.differentials)

    reverse_with_vel = get_offset_sun_from_barycenter(time, reverse=True, include_velocity=True)
    assert_quantity_allclose(reverse_with_vel.xyz, -pos.xyz)
    assert_quantity_allclose(reverse_with_vel.differentials['s'].d_xyz, -vel.xyz)



# Licensed under a 3-clause BSD style license - see LICENSE.rst

import pytest

import numpy as np

from astropy import wcs

from . helper import SimModelTAB


@pytest.fixture(scope='module')
def tab_wcs_2di():
    model = SimModelTAB(nx=150, ny=200)

    # generate FITS HDU list:
    hdulist = model.hdulist

    # create WCS object:
    w = wcs.WCS(hdulist[0].header, hdulist)

    return w


@pytest.fixture(scope='module')
def tab_wcsh_2di():
    model = SimModelTAB(nx=150, ny=200)

    # generate FITS HDU list:
    hdulist = model.hdulist

    # create WCS object:
    w = wcs.WCS(hdulist[0].header, hdulist)

    return w, hdulist


@pytest.fixture(scope='function')
def tab_wcs_2di_f():
    model = SimModelTAB(nx=150, ny=200)

    # generate FITS HDU list:
    hdulist = model.hdulist

    # create WCS object:
    w = wcs.WCS(hdulist[0].header, hdulist)

    return w



# -*- coding: utf-8 -*-

"""Cosmology classes and instances for package ``mypackage``."""

# STDLIB
from collections import UserDict

# THIRD PARTY
import numpy as np


class MyCosmology(UserDict):
    """Cosmology, from ``mypackage``.

    Parameters
    ----------
    **kw
        values for `MyCosmology`.
    """

    def __init__(self, **kw):
        super().__init__()
        self.update(kw)

    def __getattr__(self, key):
        """Get attributes like items."""
        if key in self:
            return self[key]
        super().__getattribute__(key)

    def age_in_Gyr(z):
        """Cosmology age in Gyr. Returns `NotImplemented`."""
        return NotImplemented

    def __eq__(self, other):
        """Equality check."""
        if not isinstance(other, MyCosmology):
            return NotImplemented
        return all(np.all(v == other[k]) for k, v in self.items())


myplanck = MyCosmology(
    name="planck",
    hubble_parameter=67.66,  # km/s/Mpc
    initial_dm_density=0.2607,  # X/rho_critical
    initial_baryon_density=0.04897,  # X/rho_critical
    initial_matter_density=0.30966,  # X/rho_critical
    n=0.9665,
    sigma8=0.8102,
    z_reionization=7.82,  # redshift
    current_age=13.787,  # Gyr
    initial_temperature=2.7255,  # Kelvin
    Neff=3.046,
    neutrino_masses=[0., 0., 0.06],  # eV
)



# Licensed under a 3-clause BSD style license - see LICENSE.rst

import os
from unittest.mock import patch

import pytest
import matplotlib.pyplot as plt
from astropy.wcs import WCS
from astropy.io import fits
from astropy.utils.data import get_pkg_data_filename

from astropy.visualization.wcsaxes.core import WCSAxes
from astropy import units as u

MSX_HEADER = fits.Header.fromtextfile(get_pkg_data_filename('data/msx_header'))


def teardown_function(function):
    plt.close('all')


def test_getaxislabel(ignore_matplotlibrc):

    fig = plt.figure()
    ax = WCSAxes(fig, [0.1, 0.1, 0.8, 0.8], aspect='equal')

    ax.coords[0].set_axislabel("X")
    ax.coords[1].set_axislabel("Y")
    assert ax.coords[0].get_axislabel() == "X"
    assert ax.coords[1].get_axislabel() == "Y"


@pytest.fixture
def ax():
    fig = plt.figure()
    ax = WCSAxes(fig, [0.1, 0.1, 0.8, 0.8], aspect='equal')
    fig.add_axes(ax)

    return ax


def assert_label_draw(ax, x_label, y_label):
    ax.coords[0].set_axislabel("Label 1")
    ax.coords[1].set_axislabel("Label 2")

    with patch.object(ax.coords[0].axislabels, 'set_position') as pos1:
        with patch.object(ax.coords[1].axislabels, 'set_position') as pos2:
            ax.figure.canvas.draw()

    assert pos1.call_count == x_label
    assert pos2.call_count == y_label


def test_label_visibility_rules_default(ignore_matplotlibrc, ax):
    assert_label_draw(ax, True, True)


def test_label_visibility_rules_label(ignore_matplotlibrc, ax):

    ax.coords[0].set_ticklabel_visible(False)
    ax.coords[1].set_ticks(values=[-9999]*u.one)

    assert_label_draw(ax, False, False)


def test_label_visibility_rules_ticks(ignore_matplotlibrc, ax):

    ax.coords[0].set_axislabel_visibility_rule('ticks')
    ax.coords[1].set_axislabel_visibility_rule('ticks')

    ax.coords[0].set_ticklabel_visible(False)
    ax.coords[1].set_ticks(values=[-9999]*u.one)

    assert_label_draw(ax, True, False)


def test_label_visibility_rules_always(ignore_matplotlibrc, ax):

    ax.coords[0].set_axislabel_visibility_rule('always')
    ax.coords[1].set_axislabel_visibility_rule('always')

    ax.coords[0].set_ticklabel_visible(False)
    ax.coords[1].set_ticks(values=[-9999]*u.one)

    assert_label_draw(ax, True, True)


def test_format_unit(tmpdir):

    fig = plt.figure()
    ax = WCSAxes(fig, [0.1, 0.1, 0.8, 0.8], wcs=WCS(MSX_HEADER))
    fig.add_axes(ax)

    # Force a draw which is required for format_coord to work
    ax.figure.canvas.draw()

    ori_fu = ax.coords[1].get_format_unit()
    assert ori_fu == "deg"

    ax.coords[1].set_format_unit("arcsec")
    fu = ax.coords[1].get_format_unit()
    assert fu == "arcsec"


def test_set_separator(tmpdir):

    fig = plt.figure()
    ax = WCSAxes(fig, [0.1, 0.1, 0.8, 0.8], wcs=WCS(MSX_HEADER))
    fig.add_axes(ax)

    # Force a draw which is required for format_coord to work
    ax.figure.canvas.draw()

    ax.coords[1].set_format_unit('deg')
    assert ax.coords[1].format_coord(4) == '4\xb000\'00\"'
    ax.coords[1].set_separator((':', ':', ''))
    assert ax.coords[1].format_coord(4) == '4:00:00'
    ax.coords[1].set_separator('abc')
    assert ax.coords[1].format_coord(4) == '4a00b00c'
    ax.coords[1].set_separator(None)
    assert ax.coords[1].format_coord(4) == '4\xb000\'00\"'



# coding: utf-8
# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""
Test Structured units and quantities.
"""
import copy

import pytest
import numpy as np
from numpy.testing import assert_array_equal

from astropy import units as u
from astropy.units import StructuredUnit, Unit, UnitBase, Quantity
from astropy.tests.helper import pickle_protocol, check_pickling_recovery
from astropy.utils.compat import NUMPY_LT_1_21_1
from astropy.utils.masked import Masked


class StructuredTestBase:
    @classmethod
    def setup_class(self):
        self.pv_dtype = np.dtype([('p', 'f8'), ('v', 'f8')])
        self.pv_t_dtype = np.dtype([('pv', self.pv_dtype), ('t', 'f8')])
        self.p_unit = u.km
        self.v_unit = u.km / u.s
        self.t_unit = u.s
        self.pv_dtype = np.dtype([('p', 'f8'), ('v', 'f8')])
        self.pv_t_dtype = np.dtype([('pv', self.pv_dtype), ('t', 'f8')])
        self.pv = np.array([(1., 0.25), (2., 0.5), (3., 0.75)],
                           self.pv_dtype)
        self.pv_t = np.array([((4., 2.5), 0.),
                              ((5., 5.0), 1.),
                              ((6., 7.5), 2.)], self.pv_t_dtype)


class StructuredTestBaseWithUnits(StructuredTestBase):
    @classmethod
    def setup_class(self):
        super().setup_class()
        self.pv_unit = StructuredUnit((self.p_unit, self.v_unit),
                                      ('p', 'v'))
        self.pv_t_unit = StructuredUnit((self.pv_unit, self.t_unit),
                                        ('pv', 't'))


class TestStructuredUnitBasics(StructuredTestBase):

    def test_initialization_and_keying(self):
        su = StructuredUnit((self.p_unit, self.v_unit), ('p', 'v'))
        assert su['p'] is self.p_unit
        assert su['v'] is self.v_unit
        su2 = StructuredUnit((su, self.t_unit), ('pv', 't'))
        assert isinstance(su2['pv'], StructuredUnit)
        assert su2['pv']['p'] is self.p_unit
        assert su2['pv']['v'] is self.v_unit
        assert su2['t'] is self.t_unit
        assert su2['pv'] == su
        su3 = StructuredUnit(('AU', 'AU/day'), ('p', 'v'))
        assert isinstance(su3['p'], UnitBase)
        assert isinstance(su3['v'], UnitBase)
        su4 = StructuredUnit('AU, AU/day', ('p', 'v'))
        assert su4['p'] == u.AU
        assert su4['v'] == u.AU / u.day
        su5 = StructuredUnit(('AU', 'AU/day'))
        assert su5.field_names == ('f0', 'f1')
        assert su5['f0'] == u.AU
        assert su5['f1'] == u.AU / u.day

    def test_recursive_initialization(self):
        su = StructuredUnit(((self.p_unit, self.v_unit), self.t_unit),
                            (('p', 'v'), 't'))
        assert isinstance(su['pv'], StructuredUnit)
        assert su['pv']['p'] is self.p_unit
        assert su['pv']['v'] is self.v_unit
        assert su['t'] is self.t_unit
        su2 = StructuredUnit(((self.p_unit, self.v_unit), self.t_unit),
                             (['p_v', ('p', 'v')], 't'))
        assert isinstance(su2['p_v'], StructuredUnit)
        assert su2['p_v']['p'] is self.p_unit
        assert su2['p_v']['v'] is self.v_unit
        assert su2['t'] is self.t_unit
        su3 = StructuredUnit((('AU', 'AU/day'), 'yr'),
                             (['p_v', ('p', 'v')], 't'))
        assert isinstance(su3['p_v'], StructuredUnit)
        assert su3['p_v']['p'] == u.AU
        assert su3['p_v']['v'] == u.AU / u.day
        assert su3['t'] == u.yr
        su4 = StructuredUnit('(AU, AU/day), yr', (('p', 'v'), 't'))
        assert isinstance(su4['pv'], StructuredUnit)
        assert su4['pv']['p'] == u.AU
        assert su4['pv']['v'] == u.AU / u.day
        assert su4['t'] == u.yr

    def test_extreme_recursive_initialization(self):
        su = StructuredUnit('(yr,(AU,AU/day,(km,(day,day))),m)',
                            ('t', ('p', 'v', ('h', ('d1', 'd2'))), 'l'))
        assert su.field_names == ('t', ['pvhd1d2',
                                        ('p', 'v',
                                         ['hd1d2',
                                          ('h',
                                           ['d1d2',
                                            ('d1', 'd2')])])], 'l')

    @pytest.mark.parametrize('names, invalid', [
        [('t', ['p', 'v']), "['p', 'v']"],
        [('t', ['pv', 'p', 'v']), "['pv', 'p', 'v']"],
        [('t', ['pv', ['p', 'v']]), "['pv', ['p', 'v']"],
        [('t', ()), "()"],
        [('t', ('p', None)), "None"],
        [('t', ['pv', ('p', '')]), "''"]])
    def test_initialization_names_invalid_list_errors(self, names, invalid):
        with pytest.raises(ValueError) as exc:
            StructuredUnit('(yr,(AU,AU/day)', names)
        assert f'invalid entry {invalid}' in str(exc)

    def test_looks_like_unit(self):
        su = StructuredUnit((self.p_unit, self.v_unit), ('p', 'v'))
        assert Unit(su) is su

    def test_initialize_with_float_dtype(self):
        su = StructuredUnit(('AU', 'AU/d'), self.pv_dtype)
        assert isinstance(su['p'], UnitBase)
        assert isinstance(su['v'], UnitBase)
        assert su['p'] == u.AU
        assert su['v'] == u.AU / u.day
        su = StructuredUnit((('km', 'km/s'), 'yr'), self.pv_t_dtype)
        assert isinstance(su['pv'], StructuredUnit)
        assert isinstance(su['pv']['p'], UnitBase)
        assert isinstance(su['t'], UnitBase)
        assert su['pv']['v'] == u.km / u.s
        su = StructuredUnit('(km, km/s), yr', self.pv_t_dtype)
        assert isinstance(su['pv'], StructuredUnit)
        assert isinstance(su['pv']['p'], UnitBase)
        assert isinstance(su['t'], UnitBase)
        assert su['pv']['v'] == u.km / u.s

    def test_initialize_with_structured_unit_for_names(self):
        su = StructuredUnit(('AU', 'AU/d'), names=('p', 'v'))
        su2 = StructuredUnit(('km', 'km/s'), names=su)
        assert su2.field_names == ('p', 'v')
        assert su2['p'] == u.km
        assert su2['v'] == u.km / u.s

    def test_initialize_single_field(self):
        su = StructuredUnit('AU', 'p')
        assert isinstance(su, StructuredUnit)
        assert isinstance(su['p'], UnitBase)
        assert su['p'] == u.AU
        su = StructuredUnit('AU')
        assert isinstance(su, StructuredUnit)
        assert isinstance(su['f0'], UnitBase)
        assert su['f0'] == u.AU

    def test_equality(self):
        su = StructuredUnit(('AU', 'AU/d'), self.pv_dtype)
        assert su == StructuredUnit(('AU', 'AU/d'), self.pv_dtype)
        assert su != StructuredUnit(('m', 'AU/d'), self.pv_dtype)
        # Names should be ignored.
        assert su == StructuredUnit(('AU', 'AU/d'))
        assert su == StructuredUnit(('AU', 'AU/d'), names=('q', 'w'))
        assert su != StructuredUnit(('m', 'm/s'))

    def test_parsing(self):
        su = Unit('AU, AU/d')
        assert isinstance(su, StructuredUnit)
        assert isinstance(su['f0'], UnitBase)
        assert isinstance(su['f1'], UnitBase)
        assert su['f0'] == u.AU
        assert su['f1'] == u.AU/u.day
        su2 = Unit('AU, AU/d, yr')
        assert isinstance(su2, StructuredUnit)
        assert su2 == StructuredUnit(('AU', 'AU/d', 'yr'))
        su2a = Unit('(AU, AU/d, yr)')
        assert isinstance(su2a, StructuredUnit)
        assert su2a == su2
        su3 = Unit('(km, km/s), yr')
        assert isinstance(su3, StructuredUnit)
        assert su3 == StructuredUnit((('km', 'km/s'), 'yr'))
        su4 = Unit('km,')
        assert isinstance(su4, StructuredUnit)
        assert su4 == StructuredUnit((u.km,))
        su5 = Unit('(m,s),')
        assert isinstance(su5, StructuredUnit)
        assert su5 == StructuredUnit(((u.m, u.s),))
        ldbody_unit = Unit('Msun, 0.5rad^2, (au, au/day)')
        assert ldbody_unit == StructuredUnit(
            (u.Msun, Unit(u.rad**2 / 2), (u.AU, u.AU / u.day)))

    def test_str(self):
        su = StructuredUnit(((u.km, u.km/u.s), u.yr))
        assert str(su) == '((km, km / s), yr)'
        assert Unit(str(su)) == su

    def test_repr(self):
        su = StructuredUnit(((u.km, u.km/u.s), u.yr))
        assert repr(su) == 'Unit("((km, km / s), yr)")'
        assert eval(repr(su)) == su


class TestStructuredUnitsCopyPickle(StructuredTestBaseWithUnits):
    def test_copy(self):
        su_copy = copy.copy(self.pv_t_unit)
        assert su_copy is not self.pv_t_unit
        assert su_copy == self.pv_t_unit
        assert su_copy._units is self.pv_t_unit._units

    def test_deepcopy(self):
        su_copy = copy.deepcopy(self.pv_t_unit)
        assert su_copy is not self.pv_t_unit
        assert su_copy == self.pv_t_unit
        assert su_copy._units is not self.pv_t_unit._units

    @pytest.mark.skipif(NUMPY_LT_1_21_1, reason="https://stackoverflow.com/q/69571643")
    def test_pickle(self, pickle_protocol):
        check_pickling_recovery(self.pv_t_unit, pickle_protocol)


class TestStructuredUnitAsMapping(StructuredTestBaseWithUnits):

    def test_len(self):
        assert len(self.pv_unit) == 2
        assert len(self.pv_t_unit) == 2

    def test_keys(self):
        slv = list(self.pv_t_unit.keys())
        assert slv == ['pv', 't']

    def test_values(self):
        values = self.pv_t_unit.values()
        assert values == (self.pv_unit, self.t_unit)

    def test_field_names(self):
        field_names = self.pv_t_unit.field_names
        assert isinstance(field_names, tuple)
        assert field_names == (['pv', ('p', 'v')], 't')

    @pytest.mark.parametrize('iterable', [list, set])
    def test_as_iterable(self, iterable):
        sl = iterable(self.pv_unit)
        assert isinstance(sl, iterable)
        assert sl == iterable(['p', 'v'])

    def test_as_dict(self):
        sd = dict(self.pv_t_unit)
        assert sd == {'pv': self.pv_unit, 't': self.t_unit}

    def test_contains(self):
        assert 'p' in self.pv_unit
        assert 'v' in self.pv_unit
        assert 't' not in self.pv_unit

    def test_setitem_fails(self):
        with pytest.raises(TypeError, match='item assignment'):
            self.pv_t_unit['t'] = u.Gyr


class TestStructuredUnitMethods(StructuredTestBaseWithUnits):
    def test_physical_type_id(self):
        pv_ptid = self.pv_unit._get_physical_type_id()
        assert len(pv_ptid) == 2
        assert pv_ptid.dtype.names == ('p', 'v')
        p_ptid = self.pv_unit['p']._get_physical_type_id()
        v_ptid = self.pv_unit['v']._get_physical_type_id()
        # Expected should be (subclass of) void, with structured object dtype.
        expected = np.array((p_ptid, v_ptid), [('p', 'O'), ('v', 'O')])[()]
        assert pv_ptid == expected
        # Names should be ignored in comparison.
        assert pv_ptid == np.array((p_ptid, v_ptid), 'O,O')[()]
        # Should be possible to address by field and by number.
        assert pv_ptid['p'] == p_ptid
        assert pv_ptid['v'] == v_ptid
        assert pv_ptid[0] == p_ptid
        assert pv_ptid[1] == v_ptid
        # More complicated version.
        pv_t_ptid = self.pv_t_unit._get_physical_type_id()
        t_ptid = self.t_unit._get_physical_type_id()
        assert pv_t_ptid == np.array((pv_ptid, t_ptid), 'O,O')[()]
        assert pv_t_ptid['pv'] == pv_ptid
        assert pv_t_ptid['t'] == t_ptid
        assert pv_t_ptid['pv'][1] == v_ptid

    def test_physical_type(self):
        pv_pt = self.pv_unit.physical_type
        assert pv_pt == np.array(('length', 'speed'), 'O,O')[()]

        pv_t_pt = self.pv_t_unit.physical_type
        assert pv_t_pt == np.array((pv_pt, 'time'), 'O,O')[()]

    def test_si(self):
        pv_t_si = self.pv_t_unit.si
        assert pv_t_si == self.pv_t_unit
        assert pv_t_si['pv']['v'].scale == 1000

    def test_cgs(self):
        pv_t_cgs = self.pv_t_unit.cgs
        assert pv_t_cgs == self.pv_t_unit
        assert pv_t_cgs['pv']['v'].scale == 100000

    def test_decompose(self):
        pv_t_decompose = self.pv_t_unit.decompose()
        assert pv_t_decompose['pv']['v'].scale == 1000

    def test_is_equivalent(self):
        assert self.pv_unit.is_equivalent(('AU', 'AU/day'))
        assert not self.pv_unit.is_equivalent('m')
        assert not self.pv_unit.is_equivalent(('AU', 'AU'))
        # Names should be ignored.
        pv_alt = StructuredUnit('m,m/s', names=('q', 'w'))
        assert pv_alt.field_names != self.pv_unit.field_names
        assert self.pv_unit.is_equivalent(pv_alt)
        # Regular units should work too.
        assert not u.m.is_equivalent(self.pv_unit)

    def test_conversion(self):
        pv1 = self.pv_unit.to(('AU', 'AU/day'), self.pv)
        assert isinstance(pv1, np.ndarray)
        assert pv1.dtype == self.pv.dtype
        assert np.all(pv1['p'] * u.AU == self.pv['p'] * self.p_unit)
        assert np.all(pv1['v'] * u.AU / u.day == self.pv['v'] * self.v_unit)
        # Names should be from value.
        su2 = StructuredUnit((self.p_unit, self.v_unit),
                             ('position', 'velocity'))
        pv2 = su2.to(('Mm', 'mm/s'), self.pv)
        assert pv2.dtype.names == ('p', 'v')
        assert pv2.dtype == self.pv.dtype
        # Check recursion.
        pv_t1 = self.pv_t_unit.to((('AU', 'AU/day'), 'Myr'), self.pv_t)
        assert isinstance(pv_t1, np.ndarray)
        assert pv_t1.dtype == self.pv_t.dtype
        assert np.all(pv_t1['pv']['p'] * u.AU ==
                      self.pv_t['pv']['p'] * self.p_unit)
        assert np.all(pv_t1['pv']['v'] * u.AU / u.day ==
                      self.pv_t['pv']['v'] * self.v_unit)
        assert np.all(pv_t1['t'] * u.Myr == self.pv_t['t'] * self.t_unit)
        # Passing in tuples should work.
        pv_t2 = self.pv_t_unit.to((('AU', 'AU/day'), 'Myr'),
                                  ((1., 0.1), 10.))
        assert pv_t2['pv']['p'] == self.p_unit.to('AU', 1.)
        assert pv_t2['pv']['v'] == self.v_unit.to('AU/day', 0.1)
        assert pv_t2['t'] == self.t_unit.to('Myr', 10.)
        pv_t3 = self.pv_t_unit.to((('AU', 'AU/day'), 'Myr'),
                                  [((1., 0.1), 10.),
                                   ((2., 0.2), 20.)])
        assert np.all(pv_t3['pv']['p'] == self.p_unit.to('AU', [1., 2.]))
        assert np.all(pv_t3['pv']['v'] == self.v_unit.to('AU/day', [0.1, 0.2]))
        assert np.all(pv_t3['t'] == self.t_unit.to('Myr', [10., 20.]))


class TestStructuredUnitArithmatic(StructuredTestBaseWithUnits):
    def test_multiplication(self):
        pv_times_au = self.pv_unit * u.au
        assert isinstance(pv_times_au, StructuredUnit)
        assert pv_times_au.field_names == ('p', 'v')
        assert pv_times_au['p'] == self.p_unit * u.AU
        assert pv_times_au['v'] == self.v_unit * u.AU
        au_times_pv = u.au * self.pv_unit
        assert au_times_pv == pv_times_au
        pv_times_au2 = self.pv_unit * 'au'
        assert pv_times_au2 == pv_times_au
        au_times_pv2 = 'AU' * self.pv_unit
        assert au_times_pv2 == pv_times_au
        with pytest.raises(TypeError):
            self.pv_unit * self.pv_unit
        with pytest.raises(TypeError):
            's,s' * self.pv_unit

    def test_division(self):
        pv_by_s = self.pv_unit / u.s
        assert isinstance(pv_by_s, StructuredUnit)
        assert pv_by_s.field_names == ('p', 'v')
        assert pv_by_s['p'] == self.p_unit / u.s
        assert pv_by_s['v'] == self.v_unit / u.s
        pv_by_s2 = self.pv_unit / 's'
        assert pv_by_s2 == pv_by_s
        with pytest.raises(TypeError):
            1. / self.pv_unit
        with pytest.raises(TypeError):
            u.s / self.pv_unit


class TestStructuredQuantity(StructuredTestBaseWithUnits):
    def test_initialization_and_keying(self):
        q_pv = Quantity(self.pv, self.pv_unit)
        q_p = q_pv['p']
        assert isinstance(q_p, Quantity)
        assert isinstance(q_p.unit, UnitBase)
        assert np.all(q_p == self.pv['p'] * self.pv_unit['p'])
        q_v = q_pv['v']
        assert isinstance(q_v, Quantity)
        assert isinstance(q_v.unit, UnitBase)
        assert np.all(q_v == self.pv['v'] * self.pv_unit['v'])
        q_pv_t = Quantity(self.pv_t, self.pv_t_unit)
        q_t = q_pv_t['t']
        assert np.all(q_t == self.pv_t['t'] * self.pv_t_unit['t'])
        q_pv2 = q_pv_t['pv']
        assert isinstance(q_pv2, Quantity)
        assert q_pv2.unit == self.pv_unit
        with pytest.raises(ValueError):
            Quantity(self.pv, self.pv_t_unit)
        with pytest.raises(ValueError):
            Quantity(self.pv_t, self.pv_unit)

    def test_initialization_with_unit_tuples(self):
        q_pv_t = Quantity(self.pv_t, (('km', 'km/s'), 's'))
        assert isinstance(q_pv_t.unit, StructuredUnit)
        assert q_pv_t.unit == self.pv_t_unit

    def test_initialization_with_string(self):
        q_pv_t = Quantity(self.pv_t, '(km, km/s), s')
        assert isinstance(q_pv_t.unit, StructuredUnit)
        assert q_pv_t.unit == self.pv_t_unit

    def test_initialization_by_multiplication_with_unit(self):
        q_pv_t = self.pv_t * self.pv_t_unit
        assert q_pv_t.unit is self.pv_t_unit
        assert np.all(q_pv_t.value == self.pv_t)
        assert not np.may_share_memory(q_pv_t, self.pv_t)
        q_pv_t2 = self.pv_t_unit * self.pv_t
        assert q_pv_t.unit is self.pv_t_unit
        # Not testing equality of structured Quantity here.
        assert np.all(q_pv_t2.value == q_pv_t.value)

    def test_initialization_by_shifting_to_unit(self):
        q_pv_t = self.pv_t << self.pv_t_unit
        assert q_pv_t.unit is self.pv_t_unit
        assert np.all(q_pv_t.value == self.pv_t)
        assert np.may_share_memory(q_pv_t, self.pv_t)

    def test_getitem(self):
        q_pv_t = Quantity(self.pv_t, self.pv_t_unit)
        q_pv_t01 = q_pv_t[:2]
        assert isinstance(q_pv_t01, Quantity)
        assert q_pv_t01.unit == q_pv_t.unit
        assert np.all(q_pv_t01['t'] == q_pv_t['t'][:2])
        q_pv_t1 = q_pv_t[1]
        assert isinstance(q_pv_t1, Quantity)
        assert q_pv_t1.unit == q_pv_t.unit
        assert q_pv_t1.shape == ()
        assert q_pv_t1['t'] == q_pv_t['t'][1]

    def test_value(self):
        q_pv_t = Quantity(self.pv_t, self.pv_t_unit)
        value = q_pv_t.value
        assert type(value) is np.ndarray
        assert np.all(value == self.pv_t)
        value1 = q_pv_t[1].value
        assert type(value1) is np.void
        assert np.all(value1 == self.pv_t[1])

    def test_conversion(self):
        q_pv = Quantity(self.pv, self.pv_unit)
        q1 = q_pv.to(('AU', 'AU/day'))
        assert isinstance(q1, Quantity)
        assert q1['p'].unit == u.AU
        assert q1['v'].unit == u.AU / u.day
        assert np.all(q1['p'] == q_pv['p'].to(u.AU))
        assert np.all(q1['v'] == q_pv['v'].to(u.AU/u.day))
        q2 = q_pv.to(self.pv_unit)
        assert q2['p'].unit == self.p_unit
        assert q2['v'].unit == self.v_unit
        assert np.all(q2['p'].value == self.pv['p'])
        assert np.all(q2['v'].value == self.pv['v'])
        assert not np.may_share_memory(q2, q_pv)
        pv1 = q_pv.to_value(('AU', 'AU/day'))
        assert type(pv1) is np.ndarray
        assert np.all(pv1['p'] == q_pv['p'].to_value(u.AU))
        assert np.all(pv1['v'] == q_pv['v'].to_value(u.AU/u.day))
        pv11 = q_pv[1].to_value(('AU', 'AU/day'))
        assert type(pv11) is np.void
        assert pv11 == pv1[1]
        q_pv_t = Quantity(self.pv_t, self.pv_t_unit)
        q2 = q_pv_t.to((('kpc', 'kpc/Myr'), 'Myr'))
        assert q2['pv']['p'].unit == u.kpc
        assert q2['pv']['v'].unit == u.kpc / u.Myr
        assert q2['t'].unit == u.Myr
        assert np.all(q2['pv']['p'] == q_pv_t['pv']['p'].to(u.kpc))
        assert np.all(q2['pv']['v'] == q_pv_t['pv']['v'].to(u.kpc/u.Myr))
        assert np.all(q2['t'] == q_pv_t['t'].to(u.Myr))

    def test_conversion_via_lshift(self):
        q_pv = Quantity(self.pv, self.pv_unit)
        q1 = q_pv << StructuredUnit(('AU', 'AU/day'))
        assert isinstance(q1, Quantity)
        assert q1['p'].unit == u.AU
        assert q1['v'].unit == u.AU / u.day
        assert np.all(q1['p'] == q_pv['p'].to(u.AU))
        assert np.all(q1['v'] == q_pv['v'].to(u.AU/u.day))
        q2 = q_pv << self.pv_unit
        assert q2['p'].unit == self.p_unit
        assert q2['v'].unit == self.v_unit
        assert np.all(q2['p'].value == self.pv['p'])
        assert np.all(q2['v'].value == self.pv['v'])
        assert np.may_share_memory(q2, q_pv)
        q_pv_t = Quantity(self.pv_t, self.pv_t_unit)
        q2 = q_pv_t << '(kpc,kpc/Myr),Myr'
        assert q2['pv']['p'].unit == u.kpc
        assert q2['pv']['v'].unit == u.kpc / u.Myr
        assert q2['t'].unit == u.Myr
        assert np.all(q2['pv']['p'] == q_pv_t['pv']['p'].to(u.kpc))
        assert np.all(q2['pv']['v'] == q_pv_t['pv']['v'].to(u.kpc/u.Myr))
        assert np.all(q2['t'] == q_pv_t['t'].to(u.Myr))

    def test_inplace_conversion(self):
        q_pv = Quantity(self.pv, self.pv_unit)
        q1 = q_pv.copy()
        q_link = q1
        q1 <<= StructuredUnit(('AU', 'AU/day'))
        assert q1 is q_link
        assert q1['p'].unit == u.AU
        assert q1['v'].unit == u.AU / u.day
        assert np.all(q1['p'] == q_pv['p'].to(u.AU))
        assert np.all(q1['v'] == q_pv['v'].to(u.AU/u.day))
        q_pv_t = Quantity(self.pv_t, self.pv_t_unit)
        q2 = q_pv_t.copy()
        q_link = q2
        q2 <<= '(kpc,kpc/Myr),Myr'
        assert q2 is q_link
        assert q2['pv']['p'].unit == u.kpc
        assert q2['pv']['v'].unit == u.kpc / u.Myr
        assert q2['t'].unit == u.Myr
        assert np.all(q2['pv']['p'] == q_pv_t['pv']['p'].to(u.kpc))
        assert np.all(q2['pv']['v'] == q_pv_t['pv']['v'].to(u.kpc/u.Myr))
        assert np.all(q2['t'] == q_pv_t['t'].to(u.Myr))

    def test_si(self):
        q_pv_t = Quantity(self.pv_t, self.pv_t_unit)
        q_pv_t_si = q_pv_t.si
        assert_array_equal(q_pv_t_si, q_pv_t.to('(m,m/s),s'))

    def test_cgs(self):
        q_pv_t = Quantity(self.pv_t, self.pv_t_unit)
        q_pv_t_cgs = q_pv_t.cgs
        assert_array_equal(q_pv_t_cgs, q_pv_t.to('(cm,cm/s),s'))

    def test_equality(self):
        q_pv = Quantity(self.pv, self.pv_unit)
        equal = q_pv == q_pv
        not_equal = q_pv != q_pv
        assert np.all(equal)
        assert not np.any(not_equal)
        equal2 = q_pv == q_pv[1]
        not_equal2 = q_pv != q_pv[1]
        assert np.all(equal2 == [False, True, False])
        assert np.all(not_equal2 != equal2)
        q1 = q_pv.to(('AU', 'AU/day'))
        # Ensure same conversion is done, by placing q1 first.
        assert np.all(q1 == q_pv)
        assert not np.any(q1 != q_pv)
        # Check different names in dtype.
        assert np.all(q1.value * u.Unit('AU, AU/day') == q_pv)
        assert not np.any(q1.value * u.Unit('AU, AU/day') != q_pv)
        assert (q_pv == 'b') is False
        assert ('b' != q_pv) is True
        q_pv_t = Quantity(self.pv_t, self.pv_t_unit)
        assert np.all((q_pv_t[2] == q_pv_t) == [False, False, True])
        assert np.all((q_pv_t[2] != q_pv_t) != [False, False, True])
        assert (q_pv == q_pv_t) is False
        assert (q_pv_t != q_pv) is True

    def test_setitem(self):
        q_pv = Quantity(self.pv, self.pv_unit)
        q_pv[1] = (2., 2.) * self.pv_unit
        assert q_pv[1].value == np.array((2., 2.), self.pv_dtype)
        q_pv[1:2] = (1., 0.5) * u.Unit('AU, AU/day')
        assert q_pv['p'][1] == 1. * u.AU
        assert q_pv['v'][1] == 0.5 * u.AU / u.day
        q_pv['v'] = 1. * u.km / u.s
        assert np.all(q_pv['v'] == 1. * u.km / u.s)
        with pytest.raises(u.UnitsError):
            q_pv[1] = (1., 1.) * u.Unit('AU, AU')
        with pytest.raises(u.UnitsError):
            q_pv['v'] = 1. * u.km
        q_pv_t = Quantity(self.pv_t, self.pv_t_unit)
        q_pv_t[1] = ((2., 2.), 3.) * self.pv_t_unit
        assert q_pv_t[1].value == np.array(((2., 2.), 3.), self.pv_t_dtype)
        q_pv_t[1:2] = ((1., 0.5), 5.) * u.Unit('(AU, AU/day), yr')
        assert q_pv_t['pv'][1] == (1., 0.5) * u.Unit('AU, AU/day')
        assert q_pv_t['t'][1] == 5. * u.yr
        q_pv_t['pv'] = (1., 0.5) * self.pv_unit
        assert np.all(q_pv_t['pv'] == (1., 0.5) * self.pv_unit)


class TestStructuredQuantityFunctions(StructuredTestBaseWithUnits):
    @classmethod
    def setup_class(self):
        super().setup_class()
        self.q_pv = self.pv << self.pv_unit
        self.q_pv_t = self.pv_t << self.pv_t_unit

    def test_empty_like(self):
        z = np.empty_like(self.q_pv)
        assert z.dtype == self.pv_dtype
        assert z.unit == self.pv_unit
        assert z.shape == self.pv.shape

    @pytest.mark.parametrize('func', [np.zeros_like, np.ones_like])
    def test_zeros_ones_like(self, func):
        z = func(self.q_pv)
        assert z.dtype == self.pv_dtype
        assert z.unit == self.pv_unit
        assert z.shape == self.pv.shape
        assert_array_equal(z, func(self.pv) << self.pv_unit)


class TestStructuredSpecificTypeQuantity(StructuredTestBaseWithUnits):
    def setup_class(self):
        super().setup_class()

        class PositionVelocity(u.SpecificTypeQuantity):
            _equivalent_unit = self.pv_unit

        self.PositionVelocity = PositionVelocity

    def test_init(self):
        pv = self.PositionVelocity(self.pv, self.pv_unit)
        assert isinstance(pv, self.PositionVelocity)
        assert type(pv['p']) is u.Quantity
        assert_array_equal(pv['p'], self.pv['p'] << self.pv_unit['p'])

        pv2 = self.PositionVelocity(self.pv, 'AU,AU/day')
        assert_array_equal(pv2['p'], self.pv['p'] << u.AU)

    def test_error_on_non_equivalent_unit(self):
        with pytest.raises(u.UnitsError):
            self.PositionVelocity(self.pv, 'AU')
        with pytest.raises(u.UnitsError):
            self.PositionVelocity(self.pv, 'AU,yr')


class TestStructuredLogUnit:
    def setup_class(self):
        self.mag_time_dtype = np.dtype([('mag', 'f8'), ('t', 'f8')])
        self.mag_time = np.array([(20., 10.), (25., 100.)], self.mag_time_dtype)

    def test_unit_initialization(self):
        mag_time_unit = StructuredUnit((u.STmag, u.s), self.mag_time_dtype)
        assert mag_time_unit['mag'] == u.STmag
        assert mag_time_unit['t'] == u.s

        mag_time_unit2 = u.Unit('mag(ST),s')
        assert mag_time_unit2 == mag_time_unit

    def test_quantity_initialization(self):
        su = u.Unit('mag(ST),s')
        mag_time = self.mag_time << su
        assert isinstance(mag_time['mag'], u.Magnitude)
        assert isinstance(mag_time['t'], u.Quantity)
        assert mag_time.unit == su
        assert_array_equal(mag_time['mag'], self.mag_time['mag'] << u.STmag)
        assert_array_equal(mag_time['t'], self.mag_time['t'] << u.s)

    def test_quantity_si(self):
        mag_time = self.mag_time << u.Unit('mag(ST),yr')
        mag_time_si = mag_time.si
        assert_array_equal(mag_time_si['mag'], mag_time['mag'].si)
        assert_array_equal(mag_time_si['t'], mag_time['t'].si)


class TestStructuredMaskedQuantity(StructuredTestBaseWithUnits):
    """Somewhat minimal tests.  Conversion is most stringent."""
    def setup_class(self):
        super().setup_class()
        self.qpv = self.pv << self.pv_unit
        self.pv_mask = np.array([(True, False),
                                 (False, False),
                                 (False, True)], [('p', bool), ('v', bool)])
        self.mpv = Masked(self.qpv, mask=self.pv_mask)

    def test_init(self):
        assert isinstance(self.mpv, Masked)
        assert isinstance(self.mpv, Quantity)
        assert_array_equal(self.mpv.unmasked, self.qpv)
        assert_array_equal(self.mpv.mask, self.pv_mask)

    def test_slicing(self):
        mp = self.mpv['p']
        assert isinstance(mp, Masked)
        assert isinstance(mp, Quantity)
        assert_array_equal(mp.unmasked, self.qpv['p'])
        assert_array_equal(mp.mask, self.pv_mask['p'])

    def test_conversion(self):
        mpv = self.mpv.to('AU,AU/day')
        assert isinstance(mpv, Masked)
        assert isinstance(mpv, Quantity)
        assert_array_equal(mpv.unmasked, self.qpv.to('AU,AU/day'))
        assert_array_equal(mpv.mask, self.pv_mask)
        assert np.all(mpv == self.mpv)

    def test_si(self):
        mpv = self.mpv.si
        assert isinstance(mpv, Masked)
        assert isinstance(mpv, Quantity)
        assert_array_equal(mpv.unmasked, self.qpv.si)
        assert_array_equal(mpv.mask, self.pv_mask)
        assert np.all(mpv == self.mpv)
